part of dart_flex;

typedef int CompareHandler(dynamic dataA, dynamic dataB);
typedef void HeaderMouseHandler(IItemRenderer header);
typedef IItemRenderer ItemRendererHandler(DataGridItemRenderer rowRenderer, DataGridColumn column, int index, Function defaultHandler);

class DataGrid extends ListBase {
  
  @event Stream<FrameworkEvent> onRendererAdded;
  @event Stream<FrameworkEvent> onRendererRemoved;
  @event Stream<FrameworkEvent> onHeaderRendererAdded;
  @event Stream<FrameworkEvent> onColumnsChanged;
  @event Stream<FrameworkEvent> onDataGridItemRendererFactoryChanged;
  @event Stream<FrameworkEvent> onListCSSClassesChanged;
  @event Stream<FrameworkEvent> onListScrollPositionChanged;
  @event Stream<FrameworkEvent> onHeaderHeightChanged;
  @event Stream<FrameworkEvent> onRowHeightChanged;
  @event Stream<FrameworkEvent> onColumnSpacingChanged;
  @event Stream<FrameworkEvent> onRowSpacingChanged;
  @event Stream<FrameworkEvent> onAutoManageScrollBarsChanged;
  @event Stream<FrameworkEvent> onUseSelectionEffectsChanged;
  @event Stream<FrameworkEvent> onRowLockIndexChanged;
  @event Stream<FrameworkEvent> onColumnLockIndexChanged;
  @event Stream<FrameworkEvent<int>> onHeaderResize;

  //---------------------------------
  //
  // Private properties
  //
  //---------------------------------

  List<IHeaderItemRenderer> _headerItemRenderers;
  
  List<IHeaderItemRenderer> get headerItemRenderers => _headerItemRenderers;

  VGroup _gridContainer;
  HGroup _headerBounds;
  Group _headerContainer;
  ListRenderer _list;
  bool _isSelectedIndexUpdateRequired = false;
  
  ListRenderer get list => _list;

  //---------------------------------
  //
  // Public properties
  //
  //---------------------------------

  //---------------------------------
  // columns
  //---------------------------------
  
  ObservableList<DataGridColumn> _columns;
  bool _isColumnsChanged = false;

  ObservableList<DataGridColumn> get columns => _columns;
  set columns(ObservableList<DataGridColumn> value) {
    if (value != _columns) {
      _columns = value;
      _isColumnsChanged = true;

      if (value != null) _streamSubscriptionManager.add(
          'data_grid_columnsChange', 
          value.listChanges.listen(_columns_collectionChangedHandler),
          flushExisting: true
      );

      notify('columnsChanged');

      invalidateLayout();
    }
  }
  
  //---------------------------------
  // dataGridItemRendererFactory
  //---------------------------------

  ItemRendererFactory _dataGridItemRendererFactory = new ItemRendererFactory<DataGridItemRenderer>(constructorMethod: DataGridItemRenderer.construct);

  ItemRendererFactory get dataGridItemRendererFactory => _dataGridItemRendererFactory;
  set dataGridItemRendererFactory(ItemRendererFactory value) {
    if (value != _dataGridItemRendererFactory) {
      _dataGridItemRendererFactory = value;
      
      invalidateProperties();
    }
  }
  
  //---------------------------------
  // listClasses
  //---------------------------------

  List<String> _listCssClasses = <String>[];

  List<String> get listCssClasses => _listCssClasses;

  set listCssClasses(List<String> value) {
    if (value != _listCssClasses) {
      _listCssClasses = value;

      notify('listCssClassesChanged');
      
      invalidateProperties();
    }
  }
  
  //---------------------------------
  // headless
  //---------------------------------

  bool _headless = false;

  bool get headless => _headless;
  set headless(bool value) {
    if (value != _headless) {
      _headless = value;
      
      if (_headerBounds != null) _headerBounds.visible = _headerBounds.includeInLayout = !value;
      if (_headerContainer != null) _headerContainer.visible = _headerContainer.includeInLayout = !value;

      if (_headerItemRenderers != null) {
        _headerItemRenderers.forEach(
          (IHeaderItemRenderer headerRenderer) => headerRenderer.visible = headerRenderer.includeInLayout = !value
        );
      }
    }
  }
  
  //---------------------------------
  // itemRendererHandler
  //---------------------------------

  ItemRendererHandler _itemRendererHandler;

  ItemRendererHandler get itemRendererHandler => _itemRendererHandler;
  set itemRendererHandler(ItemRendererHandler value) {
    if (value != _itemRendererHandler) {
      _itemRendererHandler = value;
      
      if (_list != null) _list._itemRenderers.forEach(
        (DataGridItemRenderer R) => R.refreshColumns()    
      );
    }
  }
  
  //---------------------------------
  // headerMouseOverHandler
  //---------------------------------

  HeaderMouseHandler _headerMouseOverHandler;

  HeaderMouseHandler get headerMouseOverHandler => _headerMouseOverHandler;
  set headerMouseOverHandler(HeaderMouseHandler value) {
    if (value != _headerMouseOverHandler) {
      _headerMouseOverHandler = value;

      invokeLaterSingle('invalidateHeaderHoverHandlers', _invalidateHeaderHoverHandlers);
    }
  }
  
  //---------------------------------
  // allowHeaderColumnSorting
  //---------------------------------

  bool _allowHeaderColumnSorting = false;

  bool get allowHeaderColumnSorting => _allowHeaderColumnSorting;
  set allowHeaderColumnSorting(bool value) {
    _allowHeaderColumnSorting = value;
  }
  
  //---------------------------------
  // allowMultipleSelection
  //---------------------------------
  
  @override
  set allowMultipleSelection(bool value) {
    super.allowMultipleSelection = value;
    
    invalidateProperties();
  }
  
  //---------------------------------
  // inactiveHandler
  //---------------------------------
  
  set inactiveHandler(InactiveHandler value) {
    super.inactiveHandler = value;
    
    invalidateProperties();
  }
  
  //---------------------------------
  // scrollPosition
  //---------------------------------
  
  int get scrollPosition => (_list != null) ? _list.scrollPosition : 0;
  set scrollPosition(int value) {
    if (_list != null) _list.setScrollPositionExternally(value);
  }
  
  //---------------------------------
  // headerMouseOutHandler
  //---------------------------------

  HeaderMouseHandler _headerMouseOutHandler;

  HeaderMouseHandler get headerMouseOutHandler => _headerMouseOutHandler;
  set headerMouseOutHandler(HeaderMouseHandler value) {
    if (value != _headerMouseOutHandler) {
      _headerMouseOutHandler = value;

      invokeLaterSingle('invalidateHeaderHoverHandlers', _invalidateHeaderHoverHandlers);
    }
  }

  //---------------------------------
  // headerHeight
  //---------------------------------

  int _headerHeight = 24;

  int get headerHeight => _headerHeight;
  set headerHeight(int value) {
    if (value != _headerHeight) {
      _headerHeight = value;

      notify('headerHeightChanged');

      invalidateProperties();
    }
  }

  //---------------------------------
  // rowHeight
  //---------------------------------

  int _rowHeight = 30;

  int get rowHeight => _rowHeight;
  set rowHeight(int value) {
    if (value != _rowHeight) {
      _rowHeight = value;

      notify('rowHeightChanged');

      invalidateProperties();
    }
  }

  //---------------------------------
  // columnSpacing
  //---------------------------------

  int _columnSpacing = 1;

  int get columnSpacing => _columnSpacing;
  set columnSpacing(int value) {
    if (value != _columnSpacing) {
      _columnSpacing = value;

      notify('columnSpacingChanged');

      invalidateProperties();
    }
  }

  //---------------------------------
  // rowSpacing
  //---------------------------------

  int _rowSpacing = 1;

  int get rowSpacing => _rowSpacing;
  set rowSpacing(int value) {
    if (value != _rowSpacing) {
      _rowSpacing = value;

      notify('rowSpacingChanged');

      invalidateProperties();
    }
  }
  
  //---------------------------------
  // autoManageScrollBars
  //---------------------------------

  bool _autoManageScrollBars = true;

  bool get autoManageScrollBars => _autoManageScrollBars;
  set autoManageScrollBars(bool value) {
    if (value != _autoManageScrollBars) {
      _autoManageScrollBars = value;

      notify('autoManageScrollBarsChanged');

      invalidateProperties();
    }
  }
  
  //---------------------------------
  // useSelectionEffects
  //---------------------------------
  
  bool _useSelectionEffects = true;

  bool get useSelectionEffects => _useSelectionEffects;
  set useSelectionEffects(bool value) {
    if (value != _useSelectionEffects) {
      _useSelectionEffects = value;

      notify('useSelectionEffectsChanged');

      invalidateProperties();
    }
  }
  
  //---------------------------------
  // autoScrollSelectionIntoView
  //---------------------------------

  bool _autoScrollSelectionIntoView = true;

  bool get autoScrollSelectionIntoView => _autoScrollSelectionIntoView;
  set autoScrollSelectionIntoView(bool value) {
    if (value != _autoScrollSelectionIntoView) {
      _autoScrollSelectionIntoView = value;

      invalidateProperties();
    }
  }
  
  //---------------------------------
  // autoScrollOnDataChange
  //---------------------------------

  bool _autoScrollOnDataChange = true;

  bool get autoScrollOnDataChange => _autoScrollOnDataChange;
  set autoScrollOnDataChange(bool value) {
    _autoScrollOnDataChange = value;
  }
  
  //---------------------------------
  // disableRecycling
  //---------------------------------
  
  bool _disableRecycling = false;
  
  bool get disableRecycling => _disableRecycling;
  set disableRecycling(bool value) {
    if (value != _disableRecycling) {
      _disableRecycling = value;
  
      invalidateProperties();
    }
  }
  
  //---------------------------------
  // useEvenOdd
  //---------------------------------
  
  bool _useEvenOdd = true;
  
  bool get useEvenOdd => _useEvenOdd;
  set useEvenOdd(bool value) {
    if (value != _useEvenOdd) {
      _useEvenOdd = value;
  
      invalidateProperties();
    }
  }
  
  //---------------------------------
  // rowLockIndex
  //---------------------------------
  
  int _rowLockIndex = -1;
  bool _isRowLockIndexChanged = false;
  
  int get rowLockIndex => _rowLockIndex;
  set rowLockIndex(int value) {
    if (value != _rowLockIndex) {
      _rowLockIndex = value;
      _isRowLockIndexChanged = true;
      
      notify('rowLockIndexChanged', value);
  
      invalidateProperties();
    }
  }
  
  //---------------------------------
  // columnLockIndex
  //---------------------------------
  
  int _columnLockIndex = -1;
  bool _isColumnLockIndexChanged = false;
  
  int get columnLockIndex => _columnLockIndex;
  set columnLockIndex(int value) {
    if (value != _columnLockIndex) {
      _columnLockIndex = value;
      _isColumnLockIndexChanged = true;
      
      notify('columnLockIndexChanged', value);
  
      invalidateProperties();
    }
  }

  //---------------------------------
  //
  // Constructor
  //
  //---------------------------------

  DataGrid() : super(elementId: null) {
	   _className = 'data-grid';
	   
	   _horizontalScrollPolicy = ScrollPolicy.AUTO;
	   _verticalScrollPolicy = ScrollPolicy.AUTO;
  }

  //---------------------------------
  //
  // Public methods
  //
  //---------------------------------
  
  @override
  void commitProperties() {
    super.commitProperties();
    
    if (_gridContainer != null) {
      _gridContainer.gap = 0;
      _gridContainer.percentWidth = 100.0;
      _gridContainer.percentHeight = 100.0;
      _gridContainer.className = 'data-grid-container';
    }
    
    if (_headerContainer != null) {
      final SpreadsheetLayout L = _headerContainer.layout as SpreadsheetLayout;
      
      L.gap = _columnSpacing;
      L.lockIndex = _columnLockIndex;
      
      _headerContainer.percentWidth = 100.0;
      _headerContainer.height = _headerHeight;
      _headerContainer.visible = _headerContainer.includeInLayout = !_headless;
      _headerContainer.autoSize = false;
      _headerContainer.className = 'data-grid-header-container';
    }
    
    if (_headerBounds != null) {
      _headerBounds.percentWidth = 100.0;
      _headerBounds.height = _headerHeight;
      _headerBounds.visible = _headerBounds.includeInLayout = !_headless;
      _headerBounds.autoSize = false;
      _headerBounds.className = 'data-grid-header-bounds';
    }
    
    if (_list != null) {
      _list.orientation = 'grid';
      _list.percentWidth = 100.0;
      _list.percentHeight = 100.0;
      _list.colPercentWidth = 100.0;
      _list.useEvenOdd = _useEvenOdd;
      _list.disableRecycling = _disableRecycling;
      _list.autoScrollSelectionIntoView = _autoScrollSelectionIntoView;
      _list.inactiveHandler = _inactiveHandler;
      _list.rowSpacing = _rowSpacing;
      _list.rowHeight = _rowHeight;
      _list.dataProvider = _dataProvider;
      _list.itemRendererFactory = _dataGridItemRendererFactory;
      _list.useSelectionEffects = _useSelectionEffects;
      _list.autoManageScrollBars = _autoManageScrollBars;
      _list.allowMultipleSelection = _allowMultipleSelection;
      _list.className = 'data-grid-list-renderer';
      _list.lockIndex = _rowLockIndex;
      
      if (_list._itemRenderers != null) 
        _list._itemRenderers.forEach((DataGridItemRenderer renderer) => renderer.lockIndex = _columnLockIndex);
    }

    if (
        _isColumnsChanged &&
        (_headerContainer != null)
    ) {
      _isColumnsChanged = false;

      _updateColumnsAndHeaders();
    }
    
    if (_isSelectedIndexUpdateRequired) {
      _isSelectedIndexUpdateRequired = false;
      
      selectedIndex = (_selectedItem != null) ? _dataProvider.indexOf(_selectedItem) : -1;
    }
    
    if (_isRowLockIndexChanged) {
      _isRowLockIndexChanged = false;
      
      if (_list != null) {
        _list.notify('listScrollPositionChanged');
        
        _list._updateAfterScrollPositionChanged();
      }
    }
    
    if (_isColumnLockIndexChanged) {
      _isColumnLockIndexChanged = false;
    }
  }
  
  void setScrollTarget(DataGridItemRenderer target, int offset) {
    final int low = target.y;
    final int high = target.y + offset;
    final int lowCompare = _list._scrollPosition;
    final int highCompare = _list._scrollPosition + _list._height;
    
    if (
        (low < lowCompare) ||
        (high >= highCompare)
    ) target._control.scrollIntoView(ScrollAlignment.CENTER);
  }
  
  void refreshColumns() {
    _isColumnsChanged = true;
    
    notify('columnsChanged');

    invalidateProperties();
  }

  //---------------------------------
  //
  // Protected methods
  //
  //---------------------------------
  
  @override
  void createChildren() {
    _gridContainer = new VGroup();
    _headerContainer = new Group()
      ..layout = new SpreadsheetLayout();
    
    _headerBounds = new HGroup();
    _list = new ListRenderer();
    
    _setupListListeners();
    
    _headerBounds.addComponent(_headerContainer);

    _gridContainer.addComponent(_headerBounds);
    _gridContainer.addComponent(_list);

    addComponent(_gridContainer);

    _setControl(new DivElement());
    
    invokeLaterSingle('updateScrollPolicy', _updateScrollPolicy);

    super.createChildren();
  }
  
  void _setupListListeners() {
    _list._streamSubscriptionManager.add(
        'data_grid_listScrollPositionChange', 
        _list.onListScrollPositionChanged.listen(
          (FrameworkEvent event) => notify('listScrollPositionChanged')
        )
    );
    
    _list._streamSubscriptionManager.add(
        'data_grid_listRendererAdded', 
        _list.onRendererAdded.listen(_list_rendererAddedHandler)
    );
    
    _list._streamSubscriptionManager.add(
        'data_grid_listRendererRemoved', 
        _list.onRendererRemoved.listen(_list_rendererRemovedHandler)
    );
    
    _list._streamSubscriptionManager.add(
        'data_grid_listHeaderScrollPositionChanged', 
        _list.onHeaderScrollPositionChanged.listen(_list_headerScrollChangedHandler)
    );
    
    _list._streamSubscriptionManager.add(
        'data_grid_listSelectedItemChanged', 
        _list.onSelectedItemChanged.listen(_list_selectedItemChangedHandler)
    );
  }
  
  @override
  void _updatePresentation() {
    super._updatePresentation();
    
    if (_list != null) {
      _list._updateVisibleItemRenderers(ignorePreviousIndex: true);
      
      _list.invokeLaterSingle('invalidateLayout', _list.invalidateLayout);
    }
  }

  void _removeAllElements() {
    if (_headerItemRenderers != null) _headerItemRenderers.removeRange(0, _headerItemRenderers.length);
    
    if (_headerContainer != null) _headerContainer.removeAll();
  }

  void _updateElements() {
    if (_list != null) _list.dataProvider = _dataProvider;
  }

  void _updateColumnsAndHeaders() {
    DataGridColumn column;
    IHeaderItemRenderer header;
    int i, len;

    _removeAllElements();
    
    if (_headerItemRenderers != null) _headerItemRenderers.forEach((IHeaderItemRenderer header) => _removeHeaderListeners(header));
 
    _headerItemRenderers = new List<IHeaderItemRenderer>();

    if (_columns != null) {
      len = _columns.length;
      
      for (i=0; i<len; i++) {
        column = _columns[i];
        
        if (column._isActive) {
          header = column.headerItemRendererFactory.immediateInstance()
            ..index = i
            ..visible = !_headless
            ..includeInLayout = !_headless
            ..height = _headerHeight
            ..data =  column.headerData;
          
          _setupHeaderListeners(header, column);
          
          _headerItemRenderers.add(header);

          _headerContainer.addComponent(header);
          
          notify('headerRendererAdded', header);
        }
      }

      if (
          (_list != null) &&
          (_list._itemRenderers != null)
      ) {
        _list._itemRenderers.forEach(
          (DataGridItemRenderer renderer) {
            renderer.gap = _columnSpacing;
            
            if (renderer.columns != _columns) renderer.columns = _columns;
            else renderer.refreshColumns();
          }
        );
      }
    }
    
    invalidateLayout();
  }
  
  List<IHeaderItemRenderer> _sortHandlers = <IHeaderItemRenderer>[];
  
  void _setupHeaderListeners(IHeaderItemRenderer header, DataGridColumn column) {
    header.streamSubscriptionManager.add(
        'data_grid_headerButtonClick', 
        header.onButtonClick.listen(_header_clickHandler)
    );
    
    header.streamSubscriptionManager.add(
        'data_grid_headerWidthChanged', 
        column.onWidthChanged.listen(
          (FrameworkEvent event) => invalidateLayout(true)
        )
    );
    
    header.streamSubscriptionManager.add(
        'data_grid_headerPercentWidthChanged', 
        column.onPercentWidthChanged.listen(
          (FrameworkEvent event) => invalidateLayout(true)
        )
    );
    
    header.streamSubscriptionManager.add(
        'data_grid_headerResizeHandler', 
        header.onHeaderResize.listen((FrameworkEvent<int> event) => _header_resizeHandler(column, event.relatedObject))
    );
    
    header.streamSubscriptionManager.add(
        'data_grid_headerResizeHandler', 
        header.onResizeTargetHovered.listen(_header_resizeTargetHovered)
    );
  }
  
  void _removeHeaderListeners(IHeaderItemRenderer header) => header.streamSubscriptionManager.flushAll();

  void _header_clickHandler(FrameworkEvent<HeaderData> event) {
    if (!_allowHeaderColumnSorting) return;
    
    final IHeaderItemRenderer renderer = event.currentTarget as IHeaderItemRenderer;
    IHeaderItemRenderer SH;
    
    if (renderer.lastClickEvent is MouseEvent) {
      final MouseEvent mouseEvent = renderer.lastClickEvent;
      
      if (!mouseEvent.ctrlKey) _sortHandlers.clear();
    } else _sortHandlers.clear();
    
    if (!_sortHandlers.contains(renderer)) _sortHandlers.add(renderer);
    
    final int maxLen = _sortHandlers.length;
    
    _headerItemRenderers.forEach((IHeaderItemRenderer H) => H.cssClasses = null);
    
    renderer.isSortedAsc = !renderer.isSortedAsc;
    
    if (_sortHandlers.length > 1) {
      for (int i=0; i<maxLen; i++) {
        SH = _sortHandlers[i];
        
        SH.cssClasses = <String>[SH.isSortedAsc ? 'sort-asc' : 'sort-desc', 'index_${i+1}'];
      }
    } else _sortHandlers.first.cssClasses = <String>[_sortHandlers.first.isSortedAsc ? 'sort-asc' : 'sort-desc'];
    
    presentationHandler = (dynamic a, dynamic b) {
      int i = 0, c = 0;
      
      while (c == 0 && i < maxLen) {
        IHeaderItemRenderer H = _sortHandlers[i++];
        
        if (H.sortHandler != null)
          c = H.isSortedAsc ? 
              H.sortHandler(a, b, _columns[_headerItemRenderers.indexOf(H)], H.data) :
             -H.sortHandler(a, b, _columns[_headerItemRenderers.indexOf(H)], H.data);
      }
      
      return c;
    };
  }
  
  @override
  void updateLayout() {
    if (
        (_list != null) &&
        (_columns != null)
    ) {
      DataGridColumn column;
      int i;
      int w = 0;
      int tw = 0;
      int remainingWidth = 0;
      double procCount = .0;

      for (i=0; i<_columns.length; i++) {
        column = _columns[i];
        
        if (column._isActive && column._isVisible) {
          if (column.percentWidth > .0) {
            procCount += column.percentWidth;

            tw += column.minWidth;
          } else if (column.width > .0) {
            w += column.width + ((i > 0) ? _columnSpacing : 0);
          }
        }
      }
      
      if (_headerItemRenderers != null) {
        _headerItemRenderers.forEach((IHeaderItemRenderer header) {
          column = _columns[_headerItemRenderers.indexOf(header)];
          
          if (column.width > 0) header.width = column.width;
          else header.percentWidth = column.percentWidth;
        });
        
        _headerContainer.invalidateLayout();
      }

      tw += w;

      remainingWidth = _width - w;

      remainingWidth = (remainingWidth < 0) ? 0 : remainingWidth;

      if (procCount > .0) {
        for (i=0; i<_columns.length; i++) {
          column = _columns[i];
          
          if (
              column._isActive &&
              (column.percentWidth > .0)
          ) w += max(column.minWidth, (remainingWidth * column.percentWidth ~/ procCount));
        }
      }
      
      _list.rowHeight = _rowHeight;
      _list.colWidth = w;
      
      if (_autoManageScrollBars) _list.horizontalScrollPolicy = (tw > _width) ? ScrollPolicy.AUTO : ScrollPolicy.NONE;
      else _list.horizontalScrollPolicy = _horizontalScrollPolicy;
      
      if (_headerBounds != null) {
        _headerBounds.width = w;
        _headerBounds.height = _headerHeight;
      }

      if (_headerContainer != null) {
        _headerContainer.width = w;
        _headerContainer.height = _headerHeight;
      }
    }

    super.updateLayout();
  }

  void _list_rendererAddedHandler(FrameworkEvent<DataGridItemRenderer> event) {
    final DataGridItemRenderer renderer = event.relatedObject
      ..className = 'data-grid-list-renderer-item-renderer'
      ..gap = _columnSpacing
      ..columns = _columns
      ..grid = this
      ..lockIndex = _columnLockIndex;
    
    renderer._streamSubscriptionManager.add(
        'data_grid_rendererDataPropertyChanged', 
        renderer.onDataPropertyChanged.listen(_renderer_dataPropertyChangedHandler),
        flushExisting: true
    );
    
    invalidateProperties();
    
    notify('rendererAdded', renderer);
  }
  
  void _list_rendererRemovedHandler(FrameworkEvent<BaseComponent> event) {
    if (event.relatedObject is DataGridItemRenderer) {
      final DataGridItemRenderer renderer = (event.relatedObject as DataGridItemRenderer)
          ..columns = null
          ..data = null
          ..field = null
          ..fields = null;
      
      notify('rendererRemoved', renderer);
    }
  }
  
  void _list_selectedItemChangedHandler(FrameworkEvent<dynamic> event) {
    selectedItem = event.relatedObject;
    selectedIndex = _list.selectedIndex;
  }
  
  void _list_headerScrollChangedHandler(FrameworkEvent event) {
    final String newValue = (_headerContainer.x - _list._headerScrollPosition).toString() + 'px';
    final SpreadsheetLayout L = _headerContainer.layout as SpreadsheetLayout;
    
    L.lockIndexPosition = _list._headerScrollPosition;
    
    if (_columnLockIndex > 0) _headerContainer.invalidateLayout(true);
    
    _reflowManager.invalidateCSS(_headerContainer._control, 'left', newValue);
  }

  void _columns_collectionChangedHandler(List<ListChangeRecord> changes) {
    _isColumnsChanged = true;

    invalidateProperties();
    
    invokeLaterSingle('updateSelection', _updateSelection);
  }
  
  @override
  void _updateSelection() {
    super._updateSelection();
    
    if (_list != null) {
      _list.selectedIndex = _selectedIndex;
      _list.selectedItem = _selectedItem;
    }
  }
  
  @override
  void _dataProvider_collectionChangedHandler(List<ListChangeRecord> changes) {
    super._dataProvider_collectionChangedHandler(changes);
    
    _isSelectedIndexUpdateRequired = true;
  }
  
  @override
  void _updateScrollPolicy() {
    if (_list != null) {
      _list.horizontalScrollPolicy = _horizontalScrollPolicy;
      _list.verticalScrollPolicy = _verticalScrollPolicy;
    }
  }
  
  void _invalidateHeaderHoverHandlers() {
    if (_headerItemRenderers != null) {
      _headerItemRenderers.forEach(
        (IHeaderItemRenderer header) {
          if (_headerMouseOutHandler != null) header.streamSubscriptionManager.add(
              'data_grid_headerMouseOut', 
              header.onMouseOut.listen(
                  (FrameworkEvent event) => _headerMouseOutHandler(event.currentTarget as IHeaderItemRenderer)
              ),
              flushExisting: true
          );
          else header.streamSubscriptionManager.flushIdent('data_grid_headerMouseOut');
          
          if (_headerMouseOverHandler != null) header.streamSubscriptionManager.add(
              'data_grid_headerMouseOver', 
              header.onMouseOver.listen(
                  (FrameworkEvent event) => _headerMouseOverHandler(event.currentTarget as IHeaderItemRenderer)    
              ),
              flushExisting: true
          );
          else header.streamSubscriptionManager.flushIdent('data_grid_headerMouseOver');
        }
      );
    }
  }
  
  void _renderer_dataPropertyChangedHandler(FrameworkEvent<IItemRenderer> event) {
    if (_autoScrollOnDataChange) event.relatedObject.control.scrollIntoView(ScrollAlignment.CENTER);
  }
  
  void _header_resizeHandler(DataGridColumn column, int newSize) {
    column.width += newSize;
    
    notify('headerResize', column.width);
  }
  
  void _header_resizeTargetHovered(FrameworkEvent<bool> event) {
    if (event.relatedObject) {
      _headerItemRenderers.forEach((IHeaderItemRenderer renderer) {
        if (renderer != event.currentTarget) renderer.isResizeTargetHovered = false;
      });
    }
  }
}